"""T2K Data Manager

Helpful tools to manage the T2K data on the grid.
"""

__version__ = '1.9.3'

import configuration
from configuration import _branding
import backends
import storage
import utils
import sys

if sys.argv[0].endswith('%s-config'%(_branding,)):
    # Someone is calling <brandname>-config.
    # Do not try to load any configuration, as that is what they might be trying to fix!
    pass
else:
    # Load configuration
    config = configuration.load_config()
    # Get backend according to configuration
    backend = backends.get_backend(config)

    # Get functions from backend
    ls = backend.ls
    iter_ls = backend.iter_ls
    ls_se = backend.ls_se
    iter_ls_se = backend.iter_ls_se
    is_dir = backend.is_dir
    is_dir_se = backend.is_dir_se
    replicas = backend.replicas
    is_file = backend.is_file
    is_file_se = backend.is_file_se
    exists = backend.exists
    is_online = backend.is_online
    checksum = backend.checksum
    state = backend.state
    replicate = backend.replicate
    remove = backend.remove
    rmdir = backend.rmdir
    move = backend.move
    rename = backend.rename
    get = backend.get
    put = backend.put

    # And from utils
    check_checksums = utils.check_checksums
    check_replicas = utils.check_replicas
    check_replica_states = utils.check_replica_states



"""Module handling the command line interface (CLI)."""

from six import print_
import cmd
import sh
import shlex
import argparse
import os
import posixpath
import t2kdm as dm
from t2kdm.commands import all_commands

def ls(*args, **kwargs):
    return [x.name for x in dm.iter_ls(*args, cached=True, **kwargs)]

class T2KDmCli(cmd.Cmd):
    """T2K Data Manager Command Line Interface (CLI)

    The CLI makes it possible to comfortably browse the grid files. All commands
    that are exposed as stand-alone scripts are also available in the CLI. This is
    ensured by registering the Commands in the `all_commands` list in the
    `commands` module.
    """

    intro = """Welcome to the T2K Data Manager CLI.
  ____  ___   _  _  ____  __  __       ___  __    ____
 (_  _)(__ \ ( )/ )(  _ \(  \/  )___  / __)(  )  (_  _)
   )(   / _/ |   (  )(_) ))    ((___)( (__  )(__  _)(_
  (__) (____)(_)\_)(____/(_/\/\_)     \___)(____)(____)

Type 'help' or '?' to list commands.
"""
    prompt = '(%s) '%(dm._branding)

    def __init__(self, *args, **kwargs):
        cmd.Cmd.__init__(self, *args, **kwargs)

        # Current directories for relative paths
        self.remotedir = posixpath.abspath('/')
        self.localdir = os.path.abspath(os.getcwd())

    def do_pwd(self, arg):
        """usage: pwd

        Print the current remote directory.
        """
        print_(self.remotedir)

    def do_lpwd(self, arg):
        """usage: lpwd

        Print the current local directory.
        """
        print_(self.localdir)

    def get_abs_remote_path(self, arg):
        """Return absolute remote path."""
        if posixpath.isabs(arg):
            return arg
        else:
            return posixpath.normpath(posixpath.join(self.remotedir, arg))

    def get_abs_local_path(self, arg):
        """Return absolute local path."""
        if os.path.isabs(arg):
            return arg
        else:
            return os.path.normpath(os.path.join(self.localdir, arg))

    def do_cd(self, arg):
        """usage: cd remotepath

        Change the current remote directory.
        """
        pwd = self.get_abs_remote_path(arg)
        # Let us see whether the path exists
        try:
            ls(pwd)
        except dm.backends.DoesNotExistException as e:
            print_(e)
        else:
            # And whether it is a directory
            if dm.is_dir(pwd, cached=True):
                self.remotedir = pwd
            else:
                print_("ERROR, not a directory: %s"%(pwd,))

    def do_lcd(self, arg):
        """usage: cd localpath

        Change the current local diretory.
        """
        pwd = self.get_abs_local_path(arg)
        if os.path.isdir(pwd):
            try:
                os.chdir(pwd)
            except OSError as e: # Catch permission errors
                print_(e)
            self.localdir = os.path.abspath(os.getcwd())
        else:
            print_("ERROR, no such local directory: %s"%(pwd,))

    def do_lls(self, arg):
        """usage: lls [-l] localpath

        List contents of local directory.
        """
        try:
            argv = shlex.split(arg)
        except ValueError as e: # Catch errors from bad bash syntax
            print_(e)
            return False

        try:
            print_(sh.ls('-1', *argv, _bg_exc=False), end='')
        except sh.ErrorReturnCode as e:
            print_(e.stderr, end='')

    def do_exit(self, arg):
        """Exit the CLI."""
        return True

    def do_quit(self, arg):
        """Exit the CLI."""
        return True

    def emptyline(self):
        print_()

    def completedefault(self, text, line, begidx, endidx):
        """Complete with content of current remote or local dir."""

        candidates = []

        # The built-in argument parsing is not very good.
        # Let's try our own.
        try:
            args = shlex.split(line)
        except ValueError: # Catch badly formatted strings
            args = line.split()

        if len(args) == 1:
            # Just the main command
            # Text should be empty
            search_text = ''
        else:
            search_text = args[-1]

        text_offset = len(search_text) - len(text)

        # Local commands start with 'l'.
        # Special case 'ls'
        if line[0] == 'l' and line[1] != 's':
            # Local path
            # Look further than just current dir
            searchdir, searchfile = os.path.split(search_text)
            abs_searchdir = searchdir
            if not os.path.isabs(abs_searchdir):
                abs_searchdir = os.path.join(self.localdir, abs_searchdir)
            # Get contents of dir
            for l in sh.ls(abs_searchdir, '-1', _iter=True):
                l = l.strip()
                if l.startswith(searchfile):
                    cand = os.path.join(searchdir, l)
                    if os.path.isdir(posixpath.join(abs_searchdir, l)):
                        cand += os.path.sep
                    candidates.append(cand[text_offset:])
        else:
            # Remote path
            # Look further than just current dir
            searchdir, searchfile = posixpath.split(search_text)
            abs_searchdir = searchdir
            if not posixpath.isabs(abs_searchdir):
                abs_searchdir = posixpath.join(self.remotedir, abs_searchdir)
            # Get contents of dir
            for l in ls(abs_searchdir, _iter=True):
                l = l.strip()
                if l.startswith(searchfile):
                    cand = posixpath.join(searchdir, l)
                    if dm.is_dir(posixpath.join(abs_searchdir, l), cached=True):
                        cand += posixpath.sep
                    candidates.append(cand[text_offset:])

        return candidates

# Load all commands into the CLI
# Each `do_X` method in the class is interpreted as a possible command for the CLI.
# Each `help_X` method in the class is called when `help X` is executed.
for command in all_commands:
    do_name = 'do_'+command.name
    # Since this is a method, the first argument will be the CLI instance
    # Also need to pass the command as default value of argument,
    # so it does not change when the variable `command` changes.
    do_cmd = lambda cli, arg, com=command: com.run_from_cli(arg, localdir=cli.localdir, remotedir=cli.remotedir)
    setattr(T2KDmCli, do_name, do_cmd) # Set the `do_X` attribute of the class

    help_name = 'help_'+command.name
    # Since this is a method, the first argument will be the CLI instance
    # Also need to pass the command as default value of argument,
    # so it does not change when the variable `command` changes.
    help_cmd = lambda cli, com=command: com.run_from_cli('-h')
    setattr(T2KDmCli, help_name, help_cmd) # Set the `help_X` attribute of the class

def run_cli():
    """ Start the T2K Data Manager - Command Line Interface."""

    parser = argparse.ArgumentParser(description="Starts the T2K Data Manager - Command Line Interface.")
    args = parser.parse_args()

    try:
        T2KDmCli().cmdloop()
    except KeyboardInterrupt: # Exit gracefully on CTRL-C
        print_('')

if __name__ == '__main__':
    run_cli()



"""Module handling the configuration of the Data Manager."""

from six.moves import configparser, input
from six import print_
from appdirs import AppDirs
import os
from os import path

# The software can be modified for use with other experiments
# The "branding" is used to make this easier
_branding = "t2kdm"
app_dirs = AppDirs(_branding, _branding)

default_values = {
    'backend':      'dirac',
    'basedir':      '/t2k.org',
    'location':     '/',
    'maid_config':  path.join(app_dirs.user_config_dir, 'maid.conf'),
    'blacklist':    '-',
}

descriptions = {
    'backend':      "Which backend should be used?\n"\
                    "Supported backends: dirac\n"\
                    "Legacy backends: gfal, lcg",
    'basedir':      "What base directory should be assumed for all files on the grid?",
    'location':     "What is your location?\n"\
                    "It must follow the general pattern of '/continent/country/site'.\n"\
                    "This is used to determine the closest storage element (SE) when downloading files.\n"\
                    "Examples: /europe/uk/ral\n"\
                    "          /americas/ca/triumf\n"\
                    "          /asia/jp/kek\n"\
                    "You can see the locations of all available SEs by running `%s-SEs`.\n"\
                    "If your location is not one of the SEs and you don't know the name of your site,\n"\
                    "you can use whatever name for the site. %s will check how many levels\n"\
                    "of your location are the same to judge the distance of each SE."%(_branding,_branding),
    'maid_config':  "Where the configuration file for the `%s-maid` command is stored.\n"\
                    "If you do not deal with raw data replication, don't worry about it."%(_branding,),
    'blacklist':    "Blacklist storage elements from being used automatically.\n"\
                    "They can still be specified explicitly.\n"\
                    "Provide the list as whitespace-separated list of SE names.\n"\
                    "Example: UKI-LT2-QMUL2-disk UKI-NORTHGRID-SHEF-HEP-disk",
}

class Configuration(object):
    """Class containing the actual configuration information."""

    def __init__(self, filename=None, defaults=default_values):
        """Initialise the configuration.

        Optionally load the configuration from a file.
        """

        # Load the default values
        for key, val in defaults.items():
            setattr(self, key, val)
        self.defaults = defaults

        # Load values from the provided file
        if filename is not None:
            self.load_config(filename)

    def load_config(self, filename):
        """Load configuration from a file."""

        # Create parser for config file
        parser = configparser.SafeConfigParser(self.defaults)
        parser.read(filename)

        # Get values from parser
        for key in self.defaults:
            setattr(self, key, parser.get('DEFAULT', key))

    def save_config(self, filename):
        """Load configuration from a file."""

        # Create parser for config file
        parser = configparser.SafeConfigParser(self.defaults)

        # Set values from config
        for key in self.defaults:
            parser.set('DEFAULT', key, getattr(self, key))

        # Save configuration to file
        with open(filename, 'wt') as f:
            parser.write(f)

    def ConfigError(self, item, message):
        return ConfigurationError("%s = %s: %s"%(item, getattr(self, item), message))

class ConfigurationError(Exception):
    """Error to be thrown if there is something wrong with the configuration."""
    pass

def load_config():
    """Load the standard configuration."""

    # Try different paths to find the configuration file
    for testpath in [
            path.join(os.getcwd(), '.%s.conf'%(_branding,)), # 1. ./.t2kdm.conf
            path.join(app_dirs.user_config_dir, '%s.conf'%(_branding,)), # 2. user_config_dir, on linux: ~/.config/t2kdm/t2kdm.conf
            path.join(app_dirs.site_config_dir, '%s.conf'%(_branding,)), # 2. site_config_dir, on linux: /etc/t2kdm/t2kdm.conf
            ]:
        if path.isfile(testpath):
            return Configuration(testpath, defaults=default_values)

    # Did not find any file, return default configuration
    return Configuration(defaults=default_values)

def run_configuration_wizard():
    """Run a configuration wizard to create a valid configuration file."""

    import argparse # import here because it is usually not needed by this module
    parser = argparse.ArgumentParser(description="Configure the T2K Data Manager")
    parser.add_argument('-l', '--local', action='store_true',
        help="save the configuration file in the current diractory as '.%s.conf'"%(_branding,))

    args = parser.parse_args()

    # Load current configuration
    conf = load_config()

    # Go through all items and ask user what should be used
    for key in default_values:
        current_value = getattr(conf, key)
        default_value = default_values[key]
        help_text = descriptions.pop(key, "-")
        text =  "\nConfiguration parameter: %s\n"\
                "\n"\
                "%s\n"\
                "\n"\
                "Current value: %s\n"\
                "Default value: %s\n"\
                %(key, help_text, current_value, default_value)
        print_(text)

        new_value = input('Enter new value [keep current]: ').strip()
        if new_value != '':
            setattr(conf, key, new_value)

    if args.local:
        outf = path.join(os.getcwd(), '.%s.conf'%(_branding,))
    else:
        outf = path.join(app_dirs.user_config_dir, '%s.conf'%(_branding,))

    print_("Saving configuration in %s"%(outf,))
    try:
        os.makedirs(path.dirname(outf))
    except OSError:
        pass
    conf.save_config(outf)

if __name__ == '__main__':
    run_configuration_wizard()



"""Tests for the T2K data Manager."""

import t2kdm as dm
import t2kdm.commands as cmd
import t2kdm.cli as cli
from  t2kdm import backends
from  t2kdm import storage
from  t2kdm import utils

import argparse
from six import print_
from contextlib import contextmanager
import sys, os, sh
import tempfile
import posixpath
import re

testdir = '/test/%s'%(dm._branding,)
testfiles = ['test1.txt', 'test2.txt', 'test3.txt']
testpaths = [posixpath.join(testdir, x) for x in testfiles]
testSEs = ['UKI-SOUTHGRID-RALPP-disk', 'UKI-SOUTHGRID-OX-HEP-disk', 'RAL-LCG2-T2K-tape']

@contextmanager
def no_output(redirect=True):
    stdout = sys.stdout
    stderr = sys.stderr
    with open('/dev/null', 'w') as null:
        try:
            if redirect:
                sys.stdout = null
                sys.stderr = null
            yield
        finally:
            sys.stdout = stdout
            sys.stderr = stderr

@contextmanager
def fake_argv(fake_argv):
    true_argv = sys.argv
    try:
        sys.argv = fake_argv
        yield
    finally:
        sys.sargv = true_argv

@contextmanager
def temp_dir():
    tempdir = tempfile.mkdtemp()
    try:
        yield tempdir
    finally:
        sh.rm('-r', tempdir)

def run_read_only_tests():
    print_("Testing ls...")

    entries = dm.backend.ls(testdir)
    for e in entries:
        if e.name == testfiles[0]:
            break
    else:
        raise Exception("Test file not in listing.")

    print_("Testing ls_se...")

    entries = dm.backend.ls_se(testdir, se=testSEs[0])
    for e in entries:
        if e.name == testfiles[0]:
            break
    else:
        raise Exception("Test file not in listing.")

    print_("Testing is_dir...")
    assert(dm.is_dir(testdir))

    print_("Testing is_dir_se...")
    assert(dm.is_dir_se(testdir, se=testSEs[0]))

    print_("Testing replicas...")
    for rep in dm.backend.replicas(testpaths[0]):
        if 'heplnx204.pp.rl.ac.uk' in rep:
            break
    else:
        raise Exception("Did not find expected replica.")

    print_("Testing is_file...")
    assert(dm.backend.is_file(testpaths[0]))
    assert(not dm.backend.is_file(testpaths[0]+"DNE"))

    print_("Testing is_file_se...")
    assert(dm.backend.is_file_se(testpaths[0], testSEs[0]))
    assert(not dm.backend.is_file_se(testpaths[0]+"DNE", testSEs[0]))

    print_("Testing exists...")
    assert(dm.backend.exists(rep))
    assert(not dm.backend.exists(posixpath.dirname(rep)))

    print_("Testing checksum...")
    assert(dm.backend.checksum(rep) == '529506c1')

    print_("Testing state...")
    assert('ONLINE' in dm.backend.state(rep))

    print_("Testing is_online...")
    assert(dm.backend.is_online(rep))

    print_("Testing StorageElement...")
    # Test distance calculation
    assert(storage.SEs[0].get_distance(storage.SEs[1]) < 0)
    # Test getting SE by host
    host = storage.SEs[0].host
    assert(host in storage.SE_by_host[host].get_storage_path('/test'))
    # Test storage path arithmetic
    assert(storage.SEs[0].get_logical_path(storage.SEs[0].get_storage_path('/test')) == '/test')
    # Test getting the closest SE
    assert(storage.get_closest_SE(testpaths[0]) is not None)

    print_("Testing get...")
    with temp_dir() as tempdir:
        path = testpaths[0]
        filename = os.path.join(tempdir, testfiles[0])

        # Test choosing source SE automatically
        assert(dm.backend.get(path, tempdir) == True)
        assert(os.path.isfile(filename))

        # Test tape refusal
        try:
            with no_output():
                dm.backend.get(testpaths[2], tempdir)
        except backends.BackendException as e:
            assert("Could not find" in e.args[0])
        else:
            raise Exception("Should have refused to download from tape!")

        # Test providing the source SE (RAL tape!)
        try:
            dm.backend.get(path, tempdir, source=testSEs[2], force=False)
        except backends.BackendException as e:
            assert("already exist" in e.args[0])
        else:
            raise Exception("Should have refused to overwrite!")
        assert(dm.backend.get(path, tempdir, source=testSEs[2], force=True) == True)
        assert(os.path.isfile(filename))
        os.remove(filename)

        # Test recursive get
        assert(dm.interactive.get(testdir, tempdir, recursive='test[12]\.txt') == 0)
        assert(os.path.isfile(filename))

    print_("Testing check...")
    with temp_dir() as tempdir:
        filename = os.path.join(tempdir, 'faulty.txt')
        with no_output(True):
            assert(dm.interactive.check(testdir, checksum=True, se=testSEs, recursive=True, quiet=False, verbose=True, list=filename) != 0) # There are some deliberate failures here
        assert os.path.isfile(filename)
        assert os.path.getsize(filename) > 0
        with no_output(True):
            assert(dm.interactive.check(testdir, se=testSEs[0:1], recursivese=testSEs[0], quiet=False, verbose=True) == 0)

    print_("Testing HTML index...")
    with temp_dir() as tempdir:
        utils.html_index("/test/", tempdir)
        utils.html_index("/test/", tempdir, recursive=True)

    print_("Testing Commands...")
    with no_output(True):
        assert(cmd.ls.run_from_cli('-l /') == False)
        assert(cmd.ls.run_from_cli('/', _return=True) == 0)

        assert(cmd.ls.run_from_cli('abc') == False) # This should not work, but not throw exception
        assert(cmd.ls.run_from_cli('"abc', _return=True) != 0) # This should not work, but not throw exception
        with fake_argv(['%s-ls'%(dm._branding,), '/']):
            assert(cmd.ls.run_from_console() == 0) # This should work
        with fake_argv(['%s-cli'%(dm._branding), '/']):
            assert(cmd.ls.run_from_console() == 0) # This should work
        with fake_argv(['%s-ls'%(dm._branding), '/abcxyz']):
            assert(cmd.ls.run_from_console() != 0) # This should not work, hence the not 0 return value

        # None of the Commands should return True in the CLI
        for com in cmd.all_commands:
            assert(com.run_from_cli('') == False)

    print_("Testing CLI...")
    cli = dm.cli.T2KDmCli()
    with no_output(True):
        cli.onecmd('help ls')
        cli.onecmd('ls .')
        cli.onecmd('cd /test')
        cli.onecmd('cd ..')
        cli.onecmd('cd /abcxyz')
        cli.onecmd('lcd /')
        cli.onecmd('lcd .')
        cli.onecmd('lcd /root')
        cli.onecmd('lcd /abcxyz')
        cli.onecmd('lls .')
        cli.onecmd('lls ".')
        assert(cli.completedefault('28', 'ls nd28', 0, 0) == ['280/'])
        assert(cli.completedefault('s', 'lls us', 0, 0) == ['sr/'])
        assert(cli.completedefault('"us', 'lls "us', 0, 0) == [])

def run_read_write_tests():
    print_("Testing replicate...")
    with no_output():
        assert(dm.interactive.replicate(testdir, testSEs[1], recursive=r'^test[1]\.t.t$', verbose=True) == 0)
        assert(dm.interactive.replicate(testdir, testSEs[1], recursive=r'^test[2]\.t.t$', source=testSEs[0], verbose=True) == 0)

    print_("Testing put...")
    with temp_dir() as tempdir:
        tempf = 'thisfileshouldnotbehereforlong.txt'
        filename = os.path.join(tempdir, tempf)
        remotename = posixpath.join(testdir, tempf)
        # Make sure the file does not exist
        try:
            for SE in storage.SEs:
                dm.remove(remotename, SE.name, final=True)
        except backends.DoesNotExistException:
            pass
        # Prepare something to upload
        with open(filename, 'wt') as f:
            f.write("This is testfile #3.\n")
        assert(dm.put(filename, testdir+'/', destination=testSEs[0]))

    print_("Testing move...")
    assert(dm.move(remotename, remotename+'dir/test.txt'))
    assert(dm.move(remotename+'dir/test.txt', remotename))
    try:
        dm.move(remotename, remotename)
    except backends.BackendException as e:
        pass
    else:
        raise Exception("Moving to existing file names should not be possible.")

    print_("Testing rename...")
    # Make sure the file does not exist
    renamed = re.sub('txt', 'TXT', remotename)
    try:
        for SE in storage.SEs:
            dm.remove(renamed, SE.name, final=True)
    except backends.DoesNotExistException:
        pass
    assert(dm.rename(remotename, 'txt', 'TXT'))
    assert(dm.rename(renamed, 'TXT', 'txt'))

    print_("Testing rmdir...")
    assert(dm.rmdir(remotename+'dir/'))
    try:
        dm.rmdir(remotename+'dir/')
    except backends.DoesNotExistException:
        pass
    else:
        raise Exception("Should have failed to delete a dir that is not there.")

    print_("Testing disk SEs...")
    # Replicate test file to all SEs, to see if they all work
    for SE in storage.SEs:
        if SE.type == 'tape' or SE.is_blacklisted():
            # These SEs do not seem to cooperate
            continue
        print_(SE.name)
        assert(dm.replicate(remotename, SE.name) == True)
        assert(SE.has_replica(remotename) == True)

    print_("Testing remove...")
    with no_output():
        assert(dm.interactive.remove(testdir, testSEs[1], recursive=True) == 0) # Remove everything from SE1
    # Remove uploaded file from previous test
    try:
        # This should fail!
        for SE in storage.SEs:
            dm.remove(remotename, SE.name)
    except backends.BackendException as e:
        assert("Only one" in e.args[0])
    else:
        raise Exception("The last copy should not have been removed!")
    # With the `final` argument it should work
    try:
        for SE in storage.SEs:
            dm.remove(remotename, SE.name, final=True)
            assert(SE.has_replica(remotename) == False)
        for SE in storage.SEs:
            dm.remove(remotename, SE.name, final=True)
    except backends.DoesNotExistException:
        # The command will fail when the file no longer exists
        pass
    else:
        raise Exception("This should have raised a DoesNotExistException at some point.")

def run_tests():
    """Test the functions of the data manager."""

    parser = argparse.ArgumentParser(description="Run tests for the T2K Data Manager.")
    parser.add_argument('-w', '--write', action='store_true',
        help="do write tests. Default: read only")
    parser.add_argument('-b', '--backend', default=None,
        help="specify which backend to use")

    args = parser.parse_args()
    if args.backend is not None:
        dm.config.backend = args.backend
        dm.backend = backends.get_backend(dm.config)

    run_read_only_tests()
    if args.write:
        run_read_write_tests()

    print_("All done.")

if __name__ == '__main__':
    run_tests()



"""Module to organise storage elements."""

import posixpath
import t2kdm as dm
from six import print_

class StorageElement(object):
    """Representation of a grid storage element"""

    def __init__(self, name, host, type, location, basepath, directpath=None, broken=False):
        """Initialise StorageElement.

        `name`: Identifier for element
        `host`: Hostname of element
        `type`: Storage type of element ('tape' or 'disk')
        `location`: Location of the SE, e.g. '/europe/uk/ral'
        `basepath`: Base path for standard storage paths on element
        `directpath`: Base path for direct access of data on the storage element (optional)
        `broken`: Is the SE broken and should not be used? Equivalent to a forced blacklisting.
        """

        self.name = name
        self.host = host
        self.basepath = basepath
        if directpath is None:
            self.directpath = basepath
        else:
            self.directpath = directpath
        self.location = location
        self.type = type
        self.broken = broken

    def is_blacklisted(self):
        """Is the SE blacklisted?"""
        return self.broken or self.name in dm.config.blacklist

    def get_storage_path(self, remotepath, direct=False):
        """Generate the standard storage path for this SE from a logical file name.

        Use the "directpath" instead of the basepath if `direct` is `True`.
        """
        if remotepath[0] != '/':
            raise ValueError("Remote path needs to be absolute, not relative!")
        if direct:
            return (self.directpath + dm.config.basedir + remotepath).strip()
        else:
            return (self.basepath + dm.config.basedir + remotepath).strip()

    def get_logical_path(self, surl):
        """Try to get the logical remotepath from a surl."""
        remotepath = None
        if surl.startswith(self.basepath):
            remotepath = surl[len(self.basepath):]
        if remotepath.startswith(dm.config.basedir):
            remotepath = remotepath[len(dm.config.basedir):]
        return remotepath

    def get_distance(self, other):
        """Return the distance to another StorageElement.

        Returns a negative number. The smaller (i.e. more negative) it is,
        the closer the two SE are together.
        """

        common = posixpath.commonprefix([self.location.lower()+'/', other.location.lower()+'/'])
        # The more '/' are in the common prefix, the closer the SEs are.
        # So we can take the negative number as measure of distance.
        distance = -common.count('/')
        return distance

    def get_replica(self, remotepath, cached=False):
        """Return the replica of the file on this SM."""
        for rep in dm.replicas(remotepath, cached=cached):
            if self.host in rep:
                return rep.strip()
        # Replica not found
        return None

    def has_replica(self, remotepath, cached=False, check_dark=False):
        """Check whether the remote path is replicated on this SE.

        If `check_dark` is `True`, check the physical file location, instead of relying on the catalogue.
        """
        if not check_dark:
            return any(self.host in replica for replica in dm.replicas(remotepath, cached=cached))
        else:
            return dm.is_file_se(remotepath, self, cached=cached)

    def get_closest_SE(self, remotepath=None, tape=False, cached=False):
        """Get the storage element with the closest replica.

        If `tape` is False (default), do not return any tape SEs.
        If no `rempotepath` is provided, just return the closest SE over all.
        """
        SEs = self.get_closest_SEs(remotepath=remotepath, tape=tape, cached=cached)
        if len(SEs) >= 1:
            return SEs[0]
        else:
            return None

    def get_closest_SEs(self, remotepath=None, tape=False, cached=False):
        """Get a list of the storage element with the closest replicas.

        If `tape` is False (default), do not return any tape SEs.
        If no `rempotepath` is provided, just return the closest SE over all.
        """
        on_tape = False

        if remotepath is None:
            candidates = SEs
        else:
            candidates = []
            for rep in dm.replicas(remotepath, cached=cached):
                cand = get_SE_by_path(rep)
                if cand is None:
                    continue
                if (cand.type == 'tape') and (tape == False):
                    on_tape = True
                    continue
                candidates.append(cand)

        if len(candidates) == 0 and on_tape:
            print_("WARNING: Replica only found on tape, but tape sources are not accepted!")

        def sorter(SE):
            if SE is None:
                return 1000
            distance = self.get_distance(SE)
            if SE.type == 'tape':
                # Prefer disks over tape, even if the tape is closer by
                distance += 10
            if SE.is_blacklisted():
                # Try blacklisted SEs only as a last resort
                distance += 100
            return distance

        return sorted(candidates, key=sorter)

    def __str__(self):
        if self.broken:
            return "%s (%s) [%s] --> BROKEN! <--"%(self.name, self.host, self.location)
        elif self.is_blacklisted():
            return "%s (%s) [%s] --> BLACKLISTED <--"%(self.name, self.host, self.location)
        else:
            return "%s (%s) [%s]"%(self.name, self.host, self.location)

# Add actual SEs
SEs = [
    StorageElement('RAL-LCG2-T2K-tape',
        host = 'srm-t2k.gridpp.rl.ac.uk',
        type = 'tape',
        location = '/europe/uk/ral',
        basepath = 'srm://srm-t2k.gridpp.rl.ac.uk:8443/srm/managerv2?SFN=/castor/ads.rl.ac.uk/prod'),
    StorageElement('UKI-SOUTHGRID-RALPP-disk',
        host = 'heplnx204.pp.rl.ac.uk',
        type = 'disk',
        location = '/europe/uk/ral',
        basepath = 'srm://heplnx204.pp.rl.ac.uk:8443/srm/managerv2?SFN=/pnfs/pp.rl.ac.uk/data/t2k'),
    StorageElement('UKI-SOUTHGRID-OX-HEP-disk',
        host = 't2se01.physics.ox.ac.uk',
        type = 'disk',
        location = '/europe/uk/ox',
        basepath = 'srm://t2se01.physics.ox.ac.uk:8446/srm/managerv2?SFN=/dpm/physics.ox.ac.uk/home/t2k.org'),
    StorageElement('UKI-NORTHGRID-SHEF-HEP-disk',
        host = 'lcgse0.shef.ac.uk',
        type = 'disk',
        location = '/europe/uk/shef',
        basepath = 'srm://lcgse0.shef.ac.uk:8446/srm/managerv2?SFN=/dpm/shef.ac.uk/home/t2k.org'),
    StorageElement('UKI-NORTHGRID-LANCS-HEP-disk',
        broken = True,
        host = 'fal-pygrid-30.lancs.ac.uk',
        type = 'disk',
        location = '/europe/uk/lancs',
        basepath = 'srm://fal-pygrid-30.lancs.ac.uk:8446/srm/managerv2?SFN=/dpm/lancs.ac.uk/home/t2k.org'),
    StorageElement('UKI-NORTHGRID-MAN-HEP-disk',
        host = 'bohr3226.tier2.hep.manchester.ac.uk',
        type = 'disk',
        location = '/europe/uk/man',
        basepath = 'srm://bohr3226.tier2.hep.manchester.ac.uk:8446/srm/managerv2?SFN=/dpm/tier2.hep.manchester.ac.uk/home/t2k.org'),
    StorageElement('UKI-NORTHGRID-LIV-HEP-disk',
        host = 'hepgrid11.ph.liv.ac.uk',
        type = 'disk',
        location = '/europe/uk/liv',
        basepath = 'srm://hepgrid11.ph.liv.ac.uk:8446/srm/managerv2?SFN=/dpm/ph.liv.ac.uk/home/t2k.org'),
    StorageElement('UKI-LT2-IC-HEP-disk',
        host = 'gfe02.grid.hep.ph.ic.ac.uk',
        type = 'disk',
        location = '/europe/uk/london/ic',
        basepath = 'srm://gfe02.grid.hep.ph.ic.ac.uk:8443/srm/managerv2?SFN=/pnfs/hep.ph.ic.ac.uk/data/t2k'),
    StorageElement('UKI-LT2-QMUL2-disk',
        host = 'se03.esc.qmul.ac.uk',
        type = 'disk',
        location = '/europe/uk/london/qmul',
        directpath = 'root://xrootd.esc.qmul.ac.uk/t2k.org',
        basepath = 'srm://se03.esc.qmul.ac.uk:8444/srm/managerv2?SFN=/t2k.org'),
    StorageElement('IN2P3-CC-disk',
        broken = True,
        host = 'in2p3.fr',
        type = 'disk',
        location = '/europe/fr/in2p3',
        basepath = 'srm://polgrid4.in2p3.fr/dpm/in2p3.fr/home/t2k.org'),
    StorageElement('pic-disk',
        host = 'srm.pic.es',
        type = 'disk',
        location = '/europe/es/pic',
        basepath = 'srm://srm.pic.es:8443/srm/managerv2?SFN=/pnfs/pic.es/data/t2k.org'),
    StorageElement('CA-TRIUMF-T2K1-disk',
        host = 't2ksrm.nd280.org',
        type = 'disk',
        location = '/americas/ca/triumf',
        basepath = 'srm://t2ksrm.nd280.org:8443/srm/managerv2?SFN=/nd280data'),
    StorageElement('CA-SFU-T21-disk',
        host = 'lcg-t2kse1.sfu.computecanada.ca',
        type = 'disk',
        location = '/americas/ca/sfu',
        basepath = 'srm://lcg-t2kse1.sfu.computecanada.ca:8443/srm/managerv2?SFN=/nd280data'),
    StorageElement('JP-KEK-CRC-02-disk',
        host = 'kek2-se01.cc.kek.jp',
        type = 'disk',
        location = '/asia/jp/kek',
        basepath = 'srm://kek2-se01.cc.kek.jp:8444/srm/managerv2?SFN=/t2k.org'),
    StorageElement('JP-KEK-CRC-02-disk-old',
        broken = True,
        host = 'kek2-tmpse.cc.kek.jp',
        type = 'disk',
        location = '/asia/jp/kek',
        basepath = 'srm://kek2-tmpse.cc.kek.jp/dpm/cc.kek.jp/home/t2k.org'),
    ]

SE_by_name = {}
SE_by_host = {}

for SE in SEs:
    SE_by_name[SE.name] = SE
    SE_by_host[SE.host] = SE

def get_SE_by_path(path):
    """Return the StorageElement corresponsing to the given srm-path."""
    for SE in SEs:
        if SE.host in path:
            return SE
    return None

def get_SE(SE):
    """Get the StorageElement by all means necessary."""
    if isinstance(SE, StorageElement):
        return SE
    if SE in SE_by_name:
        return SE_by_name[SE]
    if SE in SE_by_host:
        return SE_by_host[SE]
    return get_SE_by_path(SE)


def get_closest_SEs(remotepath=None, location=None, tape=False, cached=False):
    """Get a list of the storage element with the closest replicas.

    If `tape` is False (default), do not return any tape SEs.
    If no `rempotepath` is provided, just return the closest SE over all.
    """

    if location is None:
        location = dm.config.location
        if location == '/':
            print_("WARNING:\nWARNING: Current location is '/'. Did you configure the location with `%s-config`?\nWARNING:"%(_branding,))

    # Create a pseudo SE with the correct location
    SE = StorageElement('local',
        host = 'localhost',
        type = 'disk',
        location = location,
        basepath = '/')

    return SE.get_closest_SEs(remotepath, tape=tape, cached=cached)

def get_closest_SE(remotepath=None, location=None, tape=False, cached=False):
    """Get the storage element with the closest replica.

    If `tape` is False (default), do not return any tape SEs.
    If no `rempotepath` is provided, just return the closest SE over all.
    """
    SEs = get_closest_SEs(remotepath=remotepath, tape=tape, cached=cached)
    if len(SEs) >= 1:
        return SEs[0]
    else:
        return None



"""Interactive commands for use in scripts and CLI.

See 'commands' module for descriptions of the parameters.
"""

from six import print_
import re
import t2kdm as dm
from t2kdm import storage
from t2kdm import utils
from t2kdm import backends

class InteractiveException(Exception):
    """Exception to be raised for interactive errors, e.g. an illegal user argument."""
    pass

class _recursive(object):
    """Decorator to make a function work recursively."""

    def __init__(self, iterating="Iterating over", iterated="Succesfully iterated over"):
        self.iterating = iterating
        self.iterated = iterated
        self.function = None

    def recursive_function(self, remotepath, *args, **kwargs):
        """The recursive wrapper around the original function."""
        recursive = kwargs.pop('recursive', False)
        recursive_se = kwargs.pop('recursivese', None)
        if recursive_se is not None and recursive == False:
            recursive = True
        list_file = kwargs.pop('list', None)
        if 'verbose' in kwargs:
            verbose = kwargs['verbose']
        else:
            verbose = False

        if isinstance(recursive, str):
            regex = re.compile(recursive)
            recursive = True
        else:
            regex = None

        if list_file is not None:
            list_file = open(list_file, 'wt')

        good = 0
        bad = 0
        if recursive is True:
            for path in utils.remote_iter_recursively(remotepath, regex, se=recursive_se, ignore_exceptions=True):
                if verbose:
                    print_(self.iterating + " " + path)
                try:
                    ret = self.function(path, *args, **kwargs)
                except Exception as e:
                    if not verbose:
                        # Tell the user which file failed.
                        # Only necessary if they have not already been told.
                        print_(self.iterating + " " + path + " failed.")
                    else:
                        print_("Failed.")
                    print_(e)
                    bad += 1
                    if list_file is not None:
                        list_file.write(path + '\n')
                else:
                    if ret == 0:
                        good += 1
                    else:
                        bad += 1
                        if list_file is not None:
                            list_file.write(path + '\n')
            if verbose:
                print_("%s %d files. %d files failed."%(self.iterated, good, bad))
            if list_file is not None:
                list_file.close()
            if bad == 0:
                return 0
            else:
                return 1
        else:
            ret = self.function(remotepath, *args, **kwargs)
            if list_file is not None:
                if ret != 0:
                    list_file.write(remotepath + '\n')
                list_file.close()
            return ret

    def __call__(self, function):
        self.function = function
        return self.recursive_function

def _check_path(remotepath):
    """Make sure the remotepath is absolute."""
    if not remotepath.startswith('/'):
        raise InteractiveException("Remote path must be absolute!")

def ls(remotepath, *args, **kwargs):
    """Print the contents of a directory on screen."""
    _check_path(remotepath)

    long = kwargs.pop('long', False)
    se = kwargs.pop('se', None)
    if se is None:
        entries = dm.iter_ls(remotepath, *args, **kwargs)
    else:
        entries = dm.iter_ls_se(remotepath, *args, se=se, **kwargs)
    if long:
        # Detailed listing
        for e in entries:
            print_("{mode:<11} {links:4d} {uid:5} {gid:5} {size:13d} {modified:>12} {name}".format(
                name = e.name,
                mode = e.mode,
                links = e.links,
                uid = e.uid,
                gid = e.gid,
                size = e.size,
                modified = e.modified))
    else:
        # Just the names
        for e in entries:
            print_(e.name)
    return 0

def replicas(remotepath, *args, **kwargs):
    """Print the replicas of a file on screen."""
    _check_path(remotepath)

    checksum = kwargs.pop('checksum', False)
    state = kwargs.pop('state', False)
    name = kwargs.pop('name', False)
    reps = dm.replicas(remotepath, *args, **kwargs)
    for r in reps:
        if checksum:
            print_(dm.checksum(r), end=' ')
        if state:
            print_(dm.state(r), end=' ')
        if name:
            se = dm.storage.get_SE(r)
            if se is None:
                print_('?', end=' ')
            else:
                print_(se.name, end=' ')
        print_(r)
    return 0

@_recursive("Replicating", "Replicated")
def replicate(remotepath, *args, **kwargs):
    """Replicate files to a storage element."""
    _check_path(remotepath)

    bringonline = kwargs.pop('bringonline', False)

    if bringonline:
        timeout = 2
    else:
        timeout = 60*60*6

    ret = dm.replicate(remotepath, *args, bringonline_timeout=timeout, **kwargs)
    if ret is False:
        return 1
    else:
        return 0

@_recursive("Getting", "Downloaded")
def get(remotepath, *args, **kwargs):
    """Download files."""
    _check_path(remotepath)

    bringonline = kwargs.pop('bringonline', False)

    if bringonline:
        timeout = 2
    else:
        timeout = 60*60*6

    ret = dm.get(remotepath, *args, bringonline_timeout=timeout, **kwargs)
    if ret is False:
        return 1
    else:
        return 0

def put(localpath, remotepath, *args, **kwargs):
    """Upload a file to the grid."""
    _check_path(remotepath)

    ret = dm.put(localpath, remotepath, *args, **kwargs)
    if ret:
        return 0
    else:
        return 1

@_recursive("Removing", "Removed")
def remove(remotepath, *args, **kwargs):
    """Remove a file from a given SE."""
    _check_path(remotepath)

    ret = dm.remove(remotepath, *args, **kwargs)
    if ret == True:
        return 0
    else:
        return 1

def rmdir(remotepath, *args, **kwargs):
    """Remove a directory from the catalogue."""
    _check_path(remotepath)

    ret = dm.rmdir(remotepath, *args, **kwargs)
    if ret == True:
        return 0
    else:
        return 1

@_recursive("Moving", "Moved")
def move(oldremotepath, newremotepath, *args, **kwargs):
    """Move a file to a new position."""
    _check_path(oldremotepath)
    _check_path(newremotepath)

    ret = dm.move(oldremotepath, newremotepath, *args, **kwargs)
    if ret == True:
        return 0
    else:
        return 1

@_recursive("Renaming", "Renamed")
def rename(remotepath, regex_from, regex_to, *args, **kwargs):
    """Rename a file."""
    _check_path(remotepath)

    ret = dm.rename(remotepath, regex_from, regex_to, *args, **kwargs)
    if ret == True:
        return 0
    else:
        return 1

@_recursive("Checking", "No problems detected for")
def check(remotepath, *args, **kwargs):
    """Check if everything is alright with the files."""
    _check_path(remotepath)

    verbose = kwargs.pop('verbose', False)
    quiet = kwargs.pop('quiet', False)
    ses = kwargs.pop('se', [])
    checksum = kwargs.pop('checksum', False)
    states = kwargs.pop('states', False)

    if checksum == False and len(ses) == 0 and states == False:
        raise InteractiveException("No check specified.")

    if dm.is_dir(remotepath, cached=True):
        raise InteractiveException("%s is a directory. Maybe you want to use the `--recursive` option?"%(remotepath,))

    ret = True

    if len(ses) > 0:
        if verbose:
            print_("Checking replicas...")
        ret = ret and dm.check_replicas(remotepath, ses, cached=True)
        if not ret and not quiet:
            print_("%s is not replicated on all SEs!"%(remotepath))

    if checksum:
        if verbose:
            print_("Checking checksums...")
        chk = dm.check_checksums(remotepath, cached=True)
        if not chk and not quiet:
            print_("%s has faulty checksums!"%(remotepath))
        ret = ret and chk

    if states:
        if verbose:
            print_("Checking replica states...")
        stat = dm.check_replica_states(remotepath, cached=True)
        if not stat and not quiet:
            print_("%s has faulty replica states!"%(remotepath))
        ret = ret and stat

    if ret == True:
        return 0
    else:
        return 1

@_recursive("Fixing", "Fixed")
def fix(remotepath, **kwargs):
    ret = utils.fix_all(remotepath, **kwargs)
    if ret:
        return 0
    else:
        return 1

def print_storage_elements():
    """Print all available storage elments on screen."""

    for se in storage.SEs:
        print_(se)
    return 0

def html_index(remotepath, localpath, **kwargs):
    """Create a static html index of the catalogue."""
    ret = utils.html_index(remotepath, localpath, **kwargs)
    if ret >= 0: # html_index returns the size of directory contents
        return 0
    else:
        return 1
